"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var ConfigEvaluation_1 = __importDefault(require("./ConfigEvaluation"));
var Errors_1 = require("./Errors");
var EvaluationMetadata_1 = require("./EvaluationMetadata");
var StatsigStore_1 = __importDefault(require("./StatsigStore"));
var EvaluationUtils_1 = __importDefault(require("./utils/EvaluationUtils"));
var StickyValuesStorage_1 = __importDefault(require("./utils/StickyValuesStorage"));
var js_sha256_1 = require("js-sha256");
var CONDITION_SEGMENT_COUNT = 10 * 1000;
var USER_BUCKET_COUNT = 1000;
var Evaluator = /** @class */ (function () {
    function Evaluator(options) {
        this.options = options;
        this.store = new StatsigStore_1.default(options);
    }
    Evaluator.prototype.getFeatureGate = function (gateName) {
        return this.store.getFeatureGate(gateName);
    };
    Evaluator.prototype.getConfigSpec = function (configName) {
        return this.store.getDynamicConfig(configName);
    };
    Evaluator.prototype.getConfig = function (user, configName, options) {
        var config = this.store.getDynamicConfig(configName);
        if (config === null) {
            return new ConfigEvaluation_1.default(false, '').withEvaluationDetails(EvaluationMetadata_1.EvaluationReason.Unrecognized, this.store.getLastUpdateTime());
        }
        return this.evalConfigWithPersistedValues({
            user: user,
            unhashedName: configName,
            config: config,
            userPersistedValues: options === null || options === void 0 ? void 0 : options.userPersistedValues,
        });
    };
    Evaluator.prototype.checkGate = function (user, gateName) {
        var config = this.store.getFeatureGate(gateName);
        if (config === null) {
            return new ConfigEvaluation_1.default(false, '').withEvaluationDetails(EvaluationMetadata_1.EvaluationReason.Unrecognized, this.store.getLastUpdateTime());
        }
        return this.evalConfigSpec(user, config);
    };
    Evaluator.prototype.getLayer = function (user, layerName, options) {
        var layer = this.store.getLayerConfig(layerName);
        if (layer === null) {
            return new ConfigEvaluation_1.default(false, '').withEvaluationDetails(EvaluationMetadata_1.EvaluationReason.Unrecognized, this.store.getLastUpdateTime());
        }
        return this.evalLayerWithPersistedValues({
            user: user,
            unhashedName: layerName,
            config: layer,
            userPersistedValues: options === null || options === void 0 ? void 0 : options.userPersistedValues,
        });
    };
    Evaluator.prototype.getGlobalEvaluationDetails = function () {
        return this.store.getGlobalEvaluationDetails();
    };
    Evaluator.prototype.setInitializeValues = function (initializeValues, reason) {
        this.store.setInitializeValues(initializeValues, reason);
    };
    Evaluator.prototype.evalConfigSpec = function (user, config) {
        if (config === null) {
            return new ConfigEvaluation_1.default(false, '').withEvaluationDetails(EvaluationMetadata_1.EvaluationReason.Unrecognized, this.store.getLastUpdateTime());
        }
        var evaluation = this._eval(user, config);
        return evaluation.withEvaluationDetails(this.store.getGlobalEvaluationDetails().reason, this.store.getLastUpdateTime());
    };
    Evaluator.prototype.evalAndSaveToPersistentStorage = function (args) {
        var user = args.user, unhashedName = args.unhashedName, config = args.config;
        var evaluation = this.evalConfigSpec(user, config);
        if (evaluation.is_experiment_group) {
            StickyValuesStorage_1.default.save(user, config.idType, unhashedName, evaluation.getJSONValue());
        }
        return evaluation;
    };
    Evaluator.prototype.evalAndDeleteFromPersistentStorage = function (args) {
        var user = args.user, unhashedName = args.unhashedName, config = args.config;
        StickyValuesStorage_1.default.delete(user, config.idType, unhashedName);
        return this.evalConfigSpec(user, config);
    };
    Evaluator.prototype.evalLayerWithPersistedValues = function (args) {
        var user = args.user, unhashedName = args.unhashedName, config = args.config, userPersistedValues = args.userPersistedValues;
        if (userPersistedValues == null) {
            // Remove the sticky values from persistent storage if user persisted values is not provided
            return this.evalAndDeleteFromPersistentStorage(args);
        }
        var stickyConfig = userPersistedValues[unhashedName];
        var stickyEvaluation = stickyConfig
            ? ConfigEvaluation_1.default.fromSticky(stickyConfig)
            : null;
        if (stickyEvaluation !== null) {
            if (this.allocatedExperimentExistsAndIsActive(stickyEvaluation)) {
                // Return sticky evaluation if experiment exists and is active
                return stickyEvaluation;
            }
            else {
                // Remove the sticky values from persistent storage if the allocated experiment no longer exists or is no longer active
                return this.evalAndDeleteFromPersistentStorage(args);
            }
        }
        else {
            var evaluation = this.evalConfigSpec(user, config);
            if (this.allocatedExperimentExistsAndIsActive(evaluation)) {
                if (evaluation.is_experiment_group) {
                    // If it doesn't exist and the user is in an experiment group, then save to persisted storage.
                    StickyValuesStorage_1.default.save(user, config.idType, unhashedName, evaluation.getJSONValue());
                }
            }
            else {
                // Remove the sticky values from persistent storage if the allocated experiment no longer exists or is no longer active
                StickyValuesStorage_1.default.delete(user, config.idType, unhashedName);
            }
            return evaluation;
        }
    };
    Evaluator.prototype.allocatedExperimentExistsAndIsActive = function (evaluation) {
        var delegate = evaluation.config_delegate
            ? this.store.getDynamicConfig(evaluation.config_delegate)
            : null;
        return delegate != null && delegate.isActive == true;
    };
    Evaluator.prototype.evalConfigWithPersistedValues = function (args) {
        var user = args.user, unhashedName = args.unhashedName, config = args.config, userPersistedValues = args.userPersistedValues;
        if (!config.isActive || userPersistedValues == null) {
            // Remove the sticky values from persistent storage if experiment is not active or user persisted values is not provided
            return this.evalAndDeleteFromPersistentStorage(args);
        }
        var stickyConfig = userPersistedValues[unhashedName];
        var stickyEvaluation = stickyConfig
            ? ConfigEvaluation_1.default.fromSticky(stickyConfig)
            : null;
        if (stickyEvaluation !== null) {
            return stickyEvaluation;
        }
        // If it doesn't exist and the user is in an experiment group, then save to persisted storage.
        return this.evalAndSaveToPersistentStorage(args);
    };
    Evaluator.prototype._eval = function (user, config) {
        if (!config.enabled) {
            return new ConfigEvaluation_1.default(false, 'disabled', [], config.defaultValue);
        }
        var secondary_exposures = [];
        try {
            for (var i = 0; i < config.rules.length; i++) {
                var rule = config.rules[i];
                var ruleResult = this._evalRule(user, rule);
                secondary_exposures = secondary_exposures.concat(ruleResult.secondary_exposures);
                if (ruleResult.value === true) {
                    var delegatedResult = this._evalDelegate(user, rule, secondary_exposures);
                    if (delegatedResult) {
                        return delegatedResult;
                    }
                    var passPercent = this._evalPassPercent(user, rule, config);
                    var evaluation = new ConfigEvaluation_1.default(passPercent.value, ruleResult.rule_id, secondary_exposures, passPercent.value
                        ? ruleResult.json_value
                        : config.defaultValue, config.explicitParameters, ruleResult.config_delegate);
                    evaluation.setIsExperimentGroup(ruleResult.is_experiment_group);
                    return evaluation.withGroupName(ruleResult.group_name);
                }
            }
        }
        catch (e) {
            if (e instanceof Errors_1.StatsigUnsupportedEvaluationError) {
                return new ConfigEvaluation_1.default(false, 'default', secondary_exposures, config.defaultValue, config.explicitParameters).withEvaluationDetails(EvaluationMetadata_1.EvaluationReason.Unsupported, this.store.getLastUpdateTime());
            }
            else {
                // other error, let error boundary handle this
                throw e;
            }
        }
        return new ConfigEvaluation_1.default(false, 'default', secondary_exposures, config.defaultValue, config.explicitParameters);
    };
    Evaluator.prototype._evalDelegate = function (user, rule, exposures) {
        if (rule.configDelegate == null) {
            return null;
        }
        var config = this.store.getDynamicConfig(rule.configDelegate);
        if (!config) {
            return null;
        }
        var delegatedResult = this._eval(user, config);
        delegatedResult.config_delegate = rule.configDelegate;
        delegatedResult.undelegated_secondary_exposures = exposures;
        delegatedResult.explicit_parameters = config.explicitParameters;
        delegatedResult.secondary_exposures = exposures.concat(delegatedResult.secondary_exposures);
        return delegatedResult;
    };
    Evaluator.prototype._evalPassPercent = function (user, rule, config) {
        var _a, _b;
        if (rule.passPercentage === 100) {
            return { value: true, hash: '0' };
        }
        else if (rule.passPercentage === 0) {
            return { value: false, hash: '0' };
        }
        var hash = computeUserHash(config.salt +
            '.' +
            ((_a = rule.salt) !== null && _a !== void 0 ? _a : rule.id) +
            '.' +
            ((_b = EvaluationUtils_1.default.getUnitID(user, rule.idType)) !== null && _b !== void 0 ? _b : ''));
        return {
            value: Number(hash % BigInt(CONDITION_SEGMENT_COUNT)) <
                rule.passPercentage * 100,
            hash: String(hash),
        };
    };
    Evaluator.prototype._evalRule = function (user, rule) {
        var _a;
        var secondaryExposures = [];
        var pass = true;
        for (var _i = 0, _b = rule.conditions; _i < _b.length; _i++) {
            var condition = _b[_i];
            var result = this._evalCondition(user, condition);
            if (!result.passes) {
                pass = false;
            }
            if (result.exposures) {
                secondaryExposures = secondaryExposures.concat(result.exposures);
            }
        }
        var evaluation = new ConfigEvaluation_1.default(pass, rule.id, secondaryExposures, rule.returnValue);
        evaluation.withGroupName(rule.groupName);
        evaluation.setIsExperimentGroup((_a = rule.isExperimentGroup) !== null && _a !== void 0 ? _a : false);
        return evaluation;
    };
    Evaluator.prototype._evalCondition = function (user, condition) {
        var _a, _b, _c, _d, _e, _f;
        var value = null;
        var field = condition.field;
        var target = condition.targetValue;
        var idType = condition.idType;
        switch (condition.type.toLowerCase()) {
            case 'public':
                return { passes: true };
            case 'fail_gate':
            case 'pass_gate': {
                var nestedGate = this.store.getFeatureGate(target);
                var gateResult = this.evalConfigSpec(user, nestedGate);
                value = gateResult === null || gateResult === void 0 ? void 0 : gateResult.value;
                var allExposures = (_a = gateResult === null || gateResult === void 0 ? void 0 : gateResult.secondary_exposures) !== null && _a !== void 0 ? _a : [];
                allExposures.push({
                    gate: String(target),
                    gateValue: String(value),
                    ruleID: (_b = gateResult === null || gateResult === void 0 ? void 0 : gateResult.rule_id) !== null && _b !== void 0 ? _b : '',
                });
                return {
                    passes: condition.type.toLowerCase() === 'fail_gate' ? !value : !!value,
                    exposures: allExposures,
                };
            }
            case 'ip_based':
                value = getFromUser(user, field);
                break;
            case 'ua_based':
                value = getFromUser(user, field);
                break;
            case 'user_field':
                value = getFromUser(user, field);
                break;
            case 'environment_field':
                value = getFromEnvironment(user, field);
                break;
            case 'current_time':
                value = Date.now();
                break;
            case 'user_bucket': {
                var salt = (_c = condition.additionalValues) === null || _c === void 0 ? void 0 : _c.salt;
                var userHash = computeUserHash((_d = salt + '.' + EvaluationUtils_1.default.getUnitID(user, idType)) !== null && _d !== void 0 ? _d : '');
                value = Number(userHash % BigInt(USER_BUCKET_COUNT));
                break;
            }
            case 'unit_id':
                value = EvaluationUtils_1.default.getUnitID(user, idType);
                break;
            case 'javascript': {
                var js = (_e = condition.additionalValues) === null || _e === void 0 ? void 0 : _e.javascript;
                if (js !== null) {
                    value = eval(js);
                }
                break;
            }
            default:
                throw new Errors_1.StatsigUnsupportedEvaluationError('Unsupported condition: ' + condition.type);
        }
        var op = (_f = condition.operator) === null || _f === void 0 ? void 0 : _f.toLowerCase();
        var evalResult = false;
        switch (op) {
            // numerical
            case 'gt':
                evalResult = numberCompare(function (a, b) { return a > b; })(value, target);
                break;
            case 'gte':
                evalResult = numberCompare(function (a, b) { return a >= b; })(value, target);
                break;
            case 'lt':
                evalResult = numberCompare(function (a, b) { return a < b; })(value, target);
                break;
            case 'lte':
                evalResult = numberCompare(function (a, b) { return a <= b; })(value, target);
                break;
            // version
            case 'version_gt':
                evalResult = versionCompareHelper(function (result) { return result > 0; })(value, target);
                break;
            case 'version_gte':
                evalResult = versionCompareHelper(function (result) { return result >= 0; })(value, target);
                break;
            case 'version_lt':
                evalResult = versionCompareHelper(function (result) { return result < 0; })(value, target);
                break;
            case 'version_lte':
                evalResult = versionCompareHelper(function (result) { return result <= 0; })(value, target);
                break;
            case 'version_eq':
                evalResult = versionCompareHelper(function (result) { return result === 0; })(value, target);
                break;
            case 'version_neq':
                evalResult = versionCompareHelper(function (result) { return result !== 0; })(value, target);
                break;
            // array
            case 'any':
                evalResult = arrayAny(value, target, stringCompare(true, function (a, b) { return a === b; }));
                break;
            case 'none':
                evalResult = !arrayAny(value, target, stringCompare(true, function (a, b) { return a === b; }));
                break;
            case 'any_case_sensitive':
                evalResult = arrayAny(value, target, stringCompare(false, function (a, b) { return a === b; }));
                break;
            case 'none_case_sensitive':
                evalResult = !arrayAny(value, target, stringCompare(false, function (a, b) { return a === b; }));
                break;
            // string
            case 'str_starts_with_any':
                evalResult = arrayAny(value, target, stringCompare(true, function (a, b) { return a.startsWith(b); }));
                break;
            case 'str_ends_with_any':
                evalResult = arrayAny(value, target, stringCompare(true, function (a, b) { return a.endsWith(b); }));
                break;
            case 'str_contains_any':
                evalResult = arrayAny(value, target, stringCompare(true, function (a, b) { return a.includes(b); }));
                break;
            case 'str_contains_none':
                evalResult = !arrayAny(value, target, stringCompare(true, function (a, b) { return a.includes(b); }));
                break;
            case 'str_matches':
                try {
                    if (String(value).length < 1000) {
                        evalResult = new RegExp(target).test(String(value));
                    }
                    else {
                        evalResult = false;
                    }
                }
                catch (e) {
                    evalResult = false;
                }
                break;
            // strictly equals
            case 'eq':
                evalResult = value == target;
                break;
            case 'neq':
                evalResult = value != target;
                break;
            // dates
            case 'before':
                evalResult = dateCompare(function (a, b) { return a < b; })(value, target);
                break;
            case 'after':
                evalResult = dateCompare(function (a, b) { return a > b; })(value, target);
                break;
            case 'on':
                evalResult = dateCompare(function (a, b) {
                    a === null || a === void 0 ? void 0 : a.setHours(0, 0, 0, 0);
                    b === null || b === void 0 ? void 0 : b.setHours(0, 0, 0, 0);
                    return (a === null || a === void 0 ? void 0 : a.getTime()) === (b === null || b === void 0 ? void 0 : b.getTime());
                })(value, target);
                break;
            case 'in_segment_list':
            case 'not_in_segment_list':
                throw new Errors_1.StatsigUnsupportedEvaluationError('Unsupported condition operator: ' + op);
            default:
                throw new Errors_1.StatsigUnsupportedEvaluationError('Unsupported condition operator: ' + op);
        }
        return { passes: evalResult };
    };
    return Evaluator;
}());
exports.default = Evaluator;
function computeUserHash(userHash) {
    var buffer = js_sha256_1.sha256.create().update(userHash).array();
    var ab = new ArrayBuffer(buffer.length);
    var view = new Uint8Array(ab);
    for (var ii = 0; ii < buffer.length; ii++) {
        view[ii] = buffer[ii];
    }
    var dv = new DataView(ab);
    var hash = dv.getBigUint64(0, false);
    return hash;
}
function getFromEnvironment(user, field) {
    return getParameterCaseInsensitive(user === null || user === void 0 ? void 0 : user.statsigEnvironment, field);
}
function getParameterCaseInsensitive(object, key) {
    if (object == null) {
        return undefined;
    }
    var asLowercase = key.toLowerCase();
    var keyMatch = Object.keys(object).find(function (k) { return k.toLowerCase() === asLowercase; });
    if (keyMatch === undefined) {
        return undefined;
    }
    return object[keyMatch];
}
function getFromUser(user, field) {
    var _a, _b, _c, _d, _e, _f, _g, _h, _j;
    if (typeof user !== 'object' || user == null) {
        return null;
    }
    var indexableUser = user;
    return ((_h = (_f = (_d = (_b = (_a = indexableUser[field]) !== null && _a !== void 0 ? _a : indexableUser[field.toLowerCase()]) !== null && _b !== void 0 ? _b : (_c = user === null || user === void 0 ? void 0 : user.custom) === null || _c === void 0 ? void 0 : _c[field]) !== null && _d !== void 0 ? _d : (_e = user === null || user === void 0 ? void 0 : user.custom) === null || _e === void 0 ? void 0 : _e[field.toLowerCase()]) !== null && _f !== void 0 ? _f : (_g = user === null || user === void 0 ? void 0 : user.privateAttributes) === null || _g === void 0 ? void 0 : _g[field]) !== null && _h !== void 0 ? _h : (_j = user === null || user === void 0 ? void 0 : user.privateAttributes) === null || _j === void 0 ? void 0 : _j[field.toLowerCase()]);
}
function numberCompare(fn) {
    return function (a, b) {
        if (a == null || b == null) {
            return false;
        }
        var numA = Number(a);
        var numB = Number(b);
        if (isNaN(numA) || isNaN(numB)) {
            return false;
        }
        return fn(numA, numB);
    };
}
function versionCompareHelper(fn) {
    return function (a, b) {
        var comparison = versionCompare(a, b);
        if (comparison == null) {
            return false;
        }
        return fn(comparison);
    };
}
// Compare two version strings without the extensions.
// returns -1, 0, or 1 if first is smaller than, equal to, or larger than second.
// returns false if any of the version strings is not valid.
function versionCompare(first, second) {
    if (first == null ||
        second == null ||
        typeof first !== 'string' ||
        typeof second !== 'string') {
        return null;
    }
    var version1 = removeVersionExtension(first);
    var version2 = removeVersionExtension(second);
    if (version1.length === 0 || version2.length === 0) {
        return null;
    }
    var parts1 = version1.split('.');
    var parts2 = version2.split('.');
    for (var i = 0; i < Math.max(parts1.length, parts2.length); i++) {
        if (parts1[i] === undefined) {
            parts1[i] = '0';
        }
        if (parts2[i] === undefined) {
            parts2[i] = '0';
        }
        var n1 = Number(parts1[i]);
        var n2 = Number(parts2[i]);
        if (typeof n1 !== 'number' ||
            typeof n2 !== 'number' ||
            isNaN(n1) ||
            isNaN(n2)) {
            return null;
        }
        if (n1 < n2) {
            return -1;
        }
        else if (n1 > n2) {
            return 1;
        }
    }
    return 0;
}
function removeVersionExtension(version) {
    var hyphenIndex = version.indexOf('-');
    if (hyphenIndex >= 0) {
        return version.substr(0, hyphenIndex);
    }
    return version;
}
function stringCompare(ignoreCase, fn) {
    return function (a, b) {
        if (a == null || b == null) {
            return false;
        }
        return ignoreCase
            ? fn(String(a).toLowerCase(), String(b).toLowerCase())
            : fn(String(a), String(b));
    };
}
function dateCompare(fn) {
    return function (a, b) {
        if (a == null || b == null) {
            return false;
        }
        try {
            // Try to parse into date as a string first, if not, try unixtime
            var dateA = new Date(a);
            if (isNaN(dateA.getTime())) {
                dateA = new Date(Number(a));
            }
            var dateB = new Date(b);
            if (isNaN(dateB.getTime())) {
                dateB = new Date(Number(b));
            }
            return (!isNaN(dateA.getTime()) && !isNaN(dateB.getTime()) && fn(dateA, dateB));
        }
        catch (e) {
            // malformatted input, returning false
            return false;
        }
    };
}
function arrayAny(value, array, fn) {
    if (!Array.isArray(array)) {
        return false;
    }
    for (var i = 0; i < array.length; i++) {
        if (fn(value, array[i])) {
            return true;
        }
    }
    return false;
}
